/*
 *  UnityQTPlayback.c
 *  QTPlugin
 *
 *  Created by Frank McPherson on 2/2/07.
 *  Copyright 2007 JANUS Research Group. All rights reserved.
 *  
 *  This code is for educational purposes. 
 *  No warranty or fitness for a particular purpose is express or implied.
 *
 */

#if _MSC_VER

#define EXPORT_API __declspec(dllexport) // Visual Studio needs annotating exported functions with this

// OpenGL headers
#include <windows.h>
#include <gl/gl.h>

// QuickTime headers
#include "MacTypes.h"
#include "QTML.h"
#include "Movies.h"
#include "Gestalt.h"


#else 

#define EXPORT_API // XCode doesn't need annotated exported functions so this is empty

#include <Carbon/Carbon.h>
#include <OpenGL/gl.h>
#include <CoreServices/CoreServices.h>
#include <CoreFoundation/CoreFoundation.h>
#include <QuickTime/QuickTime.h>

#endif
#include "UnityQTPlayback.h"

static Movie sMovie = NULL;
static int movieWidth = 0;
static int movieHeight = 0;
static int lastTextureID = -1;
static short movieRefNum;
static Ptr gworldPixBase;
static void *gworldBuffer;
static PixMapHandle gworldPixMap;
static GWorldPtr gworld;


extern "C"
{

	bool QTCode_DoQTInit()
	{
		OSErr error;
		long qtVersion;
		bool qtStatus = FALSE;
		
	#if _MSC_VER
		/* Initialize QuickTime Media Layer */
		error = InitializeQTML(0L);
	#endif
		
		// Is QT installed?
		error = Gestalt(gestaltQuickTime, &qtVersion);
		
		qtStatus = Gestalt(gestaltQuickTimeVersion, &qtVersion);
		// require quicktime 7
		
		/* Initialize QuickTime */
		EnterMovies();
		
		qtStatus = TRUE;
		
			return (qtStatus);
	}


	/* Open the movie pointed to at *moviePath. The width and height must match the 
	 * width and height specified for the Unity texture, otherwise the whole texture won't 
	 * show up. If someone shows me how to do texture mapping in Unity maybe I can fix that. 
	 * The textureID is the OpenGL texture ID. 
	 * There are also plenty of pixel formats I could account for but don't. This one 
	 * works for us, and needs to match what you specify in Unity, TextureFormat.RGB24 
	 *
	 * You have to call this at least once before you can call UpdateTexture() with meaningful results 
	 * You can change movies any time by calling SetMovie() again. 
	 * After you're done playing movies make sure you call DisposeResources()
	 */
	void EXPORT_API SetMovie(char *moviePath, int width, int height, int textureID) {
		Rect movieBox;
		OSType movieDataRefType;
		Handle movieDataRef = NULL;
		OSErr myErr = noErr;
		
		movieBox.top = 0;
		movieBox.left = 0;
		movieBox.bottom = height;
		movieBox.right = width;
		movieWidth = width;
		movieHeight = height;
		
		// The first time through we've got to do a bunch of housekeeping and set-up
		if (sMovie == NULL) {
			// Initialize QuickTime
			QTCode_DoQTInit();
			
			// Create an off-screen gworld where QT can draw
			gworldBuffer = calloc(width * height * 3, 1);
			QTNewGWorldFromPtr(&gworld, k24RGBPixelFormat, &movieBox, NULL, NULL, 0,
							gworldBuffer, movieBox.right * 3);
			gworldPixMap = GetGWorldPixMap(gworld);
			LockPixels(gworldPixMap);
			gworldPixBase = GetPixBaseAddr(gworldPixMap);
			
			// Bind the texture and paint it black
			lastTextureID = textureID;
			glBindTexture(GL_TEXTURE_2D, textureID);
			void *bytes = calloc(width * height, 3);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, 
						 GL_UNSIGNED_BYTE, bytes);
			free(bytes);
			
		} else {
			// 2nd and later times through, we just need to clean up the resources we were using with the old movie
			DisposeMovie(sMovie);
			CloseMovieFile(movieRefNum);
		}

		// Open the Movie
		myErr = QTNewDataReferenceFromFullPathCFString(CFStringCreateWithCString(kCFAllocatorDefault, moviePath, kCFStringEncodingASCII), kQTNativeDefaultPathStyle, 0, &movieDataRef, &movieDataRefType);
		myErr = NewMovieFromDataRef(&sMovie, newMovieActive, &movieRefNum, movieDataRef, movieDataRefType);
			
		// Tell the movie to draw to the gworld we created
		SetMovieGWorld(sMovie, gworld, NULL);
		
	}


	void EXPORT_API UpdateTexture(int textureID) {
		// Make sure we're bound to the correct textureID
		if (textureID != lastTextureID) {
			glBindTexture(GL_TEXTURE_2D, textureID);
			lastTextureID = textureID;
		}
		if (sMovie) {
			// Yield to the movie so it can draw the newest frame
			MoviesTask(sMovie, 0);
			// The image we get from QuickTime is going to be upside down from the perspective of GL. 
			// Either render your movie upside down, or use UV mapping in the scene to fix this by flipping the y axis.
			// For example you can set the Y Tiling to -1 in the material on the same object as this texture. 
			// Pass in the rendered frame as a texture
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, movieWidth, movieHeight, 0, 
						 GL_RGB, GL_UNSIGNED_BYTE, (GLvoid *)gworldPixBase);
		}
	}

	void EXPORT_API DisposeResources() {
		if (sMovie) {
			UnlockPixels(gworldPixMap);
			DisposeGWorld(gworld);
			free(gworldBuffer);
			DisposeMovie(sMovie);
			CloseMovieFile(movieRefNum);
			ExitMovies();
	#if _MSC_VER
			TerminateQTML();
	#endif
		}
	}

	void EXPORT_API StartMoviePlaying() {
		if (sMovie) {
			StartMovie(sMovie);
		}
	}

	void EXPORT_API StopMoviePlaying() {
		if (sMovie) {
			StopMovie(sMovie);
		}
	}


	bool EXPORT_API IsMovieDonePlaying() {
		if (sMovie) {
			return IsMovieDone(sMovie);
		} else {
			return false;
		}
	}
}
